@@ -1,9 +1,11 @@
-# Copyright (C) 2011 Peter Teichman
+# Copyright (C) 2013 Peter Teichman
 
 import collections
+import itertools
 import logging
+import math
+import operator
 import os
-import pprint
 import random
 import re
 import sqlite3
@@ -50,7 +52,6 @@ class Brain:
 
         self.scorer = scoring.ScorerGroup()
         self.scorer.add_scorer(1.0, scoring.CobeScorer())
-        self.scorer.add_scorer(-1.0, scoring.IdentityScorer())
 
         tokenizer_name = graph.get_info_text("tokenizer")
         if tokenizer_name == "MegaHAL":
@@ -193,7 +194,7 @@ with its two nodes"""
         if not self._learning:
             self.graph.commit()
 
-    def reply(self, text):
+    def reply(self, text, loop_ms=500, max_len=None):
         """Reply to a string of text. If the input is not already
         Unicode, it will be decoded as utf-8."""
         if type(text) != types.UnicodeType:
@@ -221,26 +222,24 @@ with its two nodes"""
         best_score = -1.0
         best_reply = None
 
-        # loop for half a second
+        # Loop for approximately loop_ms milliseconds. This can either
+        # take more (if the first reply takes a long time to generate)
+        # or less (if the _generate_replies search ends early) time,
+        # but it should stay roughly accurate.
         start = time.time()
-        end = start + 0.5
+        end = start + loop_ms * 0.001
         count = 0
 
         all_replies = []
 
         _start = time.time()
-        while time.time() < end:
-            with trace_us("Brain.generate_reply_us"):
-                candidate = self._generate_reply(pivot_set)
-
-            if candidate is None:
-                continue
-
-            count += 1
-            edges, pivot_node = candidate
+        for edges, pivot_node in self._generate_replies(pivot_set):
             reply = Reply(self.graph, tokens, input_ids, pivot_node, edges)
 
-            key = self._get_reply_key(reply)
+            if max_len and self._too_long(max_len, reply):
+                continue
+
+            key = reply.edge_ids
             if key not in score_cache:
                 with trace_us("Brain.evaluate_reply_us"):
                     score = self.scorer.score(reply)
@@ -257,6 +256,16 @@ with its two nodes"""
             if log.isEnabledFor(logging.DEBUG):
                 all_replies.append((score, reply))
 
+            count += 1
+            if time.time() > end:
+                break
+
+        if best_reply is None:
+            # we couldn't find any pivot words in _babble(), so we're
+            # working with an essentially empty brain. Use the classic
+            # MegaHAL reply:
+            return "I don't know enough to answer you yet!"
+
         _time = time.time() - _start
 
         if best_reply is None:
@@ -273,7 +282,7 @@ with its two nodes"""
             replies.sort()
 
             for score, text in replies:
-                log.debug("%f %s", score, text.encode("utf-8"))
+                log.debug("%f %s", score, text)
 
         trace("Brain.reply_input_token_count", len(tokens))
         trace("Brain.known_word_token_count", len(pivot_set))
@@ -281,10 +290,17 @@ with its two nodes"""
         trace("Brain.reply_us", _time)
         trace("Brain.reply_count", count, _time)
         trace("Brain.best_reply_score", int(best_score * 1000))
-        trace("Brain.best_reply_length", len(best_reply.edges))
+        trace("Brain.best_reply_length", len(best_reply.edge_ids))
 
-        log.debug("made %d replies (%d unique) in %f seconds" \
-                      % (count, len(score_cache), _time))
+        log.debug("made %d replies (%d unique) in %f seconds"
+                  % (count, len(score_cache), _time))
+
+        if len(text) > 60:
+            msg = text[0:60] + "..."
+        else:
+            msg = text
+
+        log.info("[%s] %d %f", msg, count, best_score)
 
         # look up the words for these tokens
         with trace_us("Brain.reply_words_lookup_us"):
@@ -292,24 +308,22 @@ with its two nodes"""
 
         return text
 
+    def _too_long(self, max_len, reply):
+        text = reply.to_text()
+        if len(text) > max_len:
+            log.debug("over max_len [%d]: %s", len(text), text)
+            return True
+
     def _conflate_stems(self, pivot_set, tokens):
         for token in tokens:
             stem_ids = self.graph.get_token_stem_id(self.stemmer.stem(token))
-            if len(stem_ids) == 0:
+            if not stem_ids:
                 continue
 
             # add the tuple of stems to the pivot set, and then
             # remove the individual token_ids
-            pivot_set.add(stem_ids)
-
-            for stem_id in stem_ids:
-                try:
-                    pivot_set.remove(stem_id)
-                except KeyError:
-                    pass
-
-    def _get_reply_key(self, reply):
-        return tuple([edge.edge_id for edge in reply.edges])
+            pivot_set.add(tuple(stem_ids))
+            pivot_set.difference_update(stem_ids)
 
     def _babble(self):
         token_ids = []
@@ -320,19 +334,19 @@ with its two nodes"""
             if token_id is not None:
                 token_ids.append(token_id)
 
-        return token_ids
+        return set(token_ids)
 
     def _filter_pivots(self, pivots):
         # remove pivots that might not give good results
         tokens = set(filter(None, pivots))
 
         filtered = self.graph.get_word_tokens(tokens)
-        if len(filtered) == 0:
-            filtered = self.graph.get_tokens(tokens)
+        if not filtered:
+            filtered = self.graph.get_tokens(tokens) or []
 
         return set(filtered)
 
-    def _choose_pivot(self, pivot_ids):
+    def _pick_pivot(self, pivot_ids):
         pivot = random.choice(tuple(pivot_ids))
 
         if type(pivot) is types.TupleType:
@@ -341,24 +355,41 @@ with its two nodes"""
 
         return pivot
 
-    def _generate_reply(self, pivot_ids):
-        if len(pivot_ids) == 0:
+    def _generate_replies(self, pivot_ids):
+        if not pivot_ids:
             return
 
-        # generate a reply containing one of token_ids
-        pivot_id = self._choose_pivot(pivot_ids)
-        node = self.graph.get_random_node_with_token(pivot_id)
+        end = self._end_context_id
+        graph = self.graph
+        search = graph.search_random_walk
 
-        if node is None:
-            return
+        # Cache all the trailing and beginning sentences we find from
+        # each random node we search. Since the node is a full n-tuple
+        # context, we can combine any pair of next_cache[node] and
+        # prev_cache[node] and get a new reply.
+        next_cache = collections.defaultdict(set)
+        prev_cache = collections.defaultdict(set)
 
-        edges = collections.deque()
+        while pivot_ids:
+            # generate a reply containing one of token_ids
+            pivot_id = self._pick_pivot(pivot_ids)
+            node = graph.get_random_node_with_token(pivot_id)
 
-        self.graph.walk(node, self._end_context_id, 1, edges.append)
-        self.graph.walk(node, self._end_context_id, 0, edges.appendleft)
+            parts = itertools.izip_longest(search(node, end, 1),
+                                           search(node, end, 0),
+                                           fillvalue=None)
 
-        if len(edges):
-            return edges, node
+            for next, prev in parts:
+                if next:
+                    next_cache[node].add(next)
+                    for p in prev_cache[node]:
+                        yield p + next, node
+
+                if prev:
+                    prev = tuple(reversed(prev))
+                    prev_cache[node].add(prev)
+                    for n in next_cache[node]:
+                        yield prev + n, node
 
     @staticmethod
     def init(filename, order=3, tokenizer=None):
@@ -385,52 +416,26 @@ tokenizer -- One of Cobe, MegaHAL (default Cobe). See documentation
 
 class Reply:
     """Provide useful support for scoring functions"""
-    def __init__(self, graph, tokens, token_ids, pivot_node, edges):
+    def __init__(self, graph, tokens, token_ids, pivot_node, edge_ids):
         self.graph = graph
         self.tokens = tokens
         self.token_ids = token_ids
         self.pivot_node = pivot_node
-        self.edges = edges
-
-    def to_graph(self):
-        text = []
-        for edge in self.edges:
-            text.append(edge.pretty())
-
-        return pprint.pformat(text)
+        self.edge_ids = edge_ids
+        self.text = None
 
     def to_text(self):
-        text = []
-        for edge in self.edges:
-            text.append(edge.get_prev_word())
-            if edge.has_space:
-                text.append(" ")
-        return "".join(text)
+        if self.text is None:
+            parts = []
+            for word, has_space in map(self.graph.get_text_by_edge,
+                                       self.edge_ids):
+                parts.append(word)
+                if has_space:
+                    parts.append(" ")
 
+            self.text = "".join(parts)
 
-class Edge:
-    def __init__(self, graph, edge_id, prev, next, has_space, count):
-        self.graph = graph
-
-        self.edge_id = edge_id
-        self.prev = prev
-        self.next = next
-        self.has_space = has_space
-        self.count = count
-
-    def get_prev_word(self):
-        # get the last word in the prev context
-        return self.graph.get_word_by_node(self.prev)
-
-    def get_prev_token(self):
-        # get the last token in the prev context
-        return self.graph.get_token_by_node(self.prev)
-
-    def pretty(self):
-        prev = "|".join(self.graph.get_node_text(self.prev))
-        next = "|".join(self.graph.get_node_text(self.next))
-
-        return "%s -> %s (%s -> %s)" % (prev, next, self.prev, self.next)
+        return self.text
 
 
 class Graph:
@@ -452,9 +457,11 @@ class Graph:
             self._all_tokens_q = ",".join(["?" for i in xrange(self.order)])
             self._last_token = "token%d_id" % (self.order - 1)
 
-            # Use a 10M cache by default. This speeds replies quite a bit.
+            # Disable the SQLite cache. Its pages tend to get swapped
+            # out, even if the database file is in buffer cache.
             c = self.cursor()
-            c.execute("PRAGMA cache_size=10000")
+            c.execute("PRAGMA cache_size=0")
+            c.execute("PRAGMA page_size=4096")
 
             # Each of these speed-for-reliability tradeoffs is useful for
             # bulk learning.
@@ -536,8 +543,10 @@ class Graph:
             c.execute(q, (text, is_word))
 
             token_id = c.lastrowid
-            if is_word and stemmer is not None:
-                self.insert_stem(token_id, stemmer.stem(text))
+            if stemmer is not None:
+                stem = stemmer.stem(text)
+                if stem is not None:
+                    self.insert_stem(token_id, stem)
 
             return token_id
 
@@ -545,35 +554,11 @@ class Graph:
         q = "INSERT INTO token_stems (token_id, stem) VALUES (?, ?)"
         self._conn.execute(q, (token_id, stem))
 
-    def get_token_by_id(self, token_id):
-        q = "SELECT text FROM tokens WHERE id = ?"
-        row = self._conn.execute(q, (token_id,)).fetchone()
-        if row:
-            return row[0]
-
     def get_token_stem_id(self, stem):
         q = "SELECT token_id FROM token_stems WHERE token_stems.stem = ?"
         rows = self._conn.execute(q, (stem,))
         if rows:
-            return tuple(val[0] for val in rows)
-
-    def get_word_by_node(self, node_id):
-        # return the last word in the node
-        q = "SELECT tokens.text FROM nodes, tokens WHERE nodes.id = ? " \
-            "AND %s = tokens.id" % self._last_token
-
-        row = self._conn.execute(q, (node_id,)).fetchone()
-        if row:
-            return row[0]
-
-    def get_token_by_node(self, node_id):
-        # return the last token in the node
-        q = "SELECT tokens.id FROM nodes, tokens WHERE nodes.id = ? " \
-            "AND %s = tokens.id" % self._last_token
-
-        row = self._conn.execute(q, (node_id,)).fetchone()
-        if row:
-            return row[0]
+            return map(operator.itemgetter(0), rows)
 
     def get_word_tokens(self, token_ids):
         q = "SELECT id FROM tokens WHERE id IN %s AND is_word = 1" % \
@@ -581,9 +566,7 @@ class Graph:
 
         rows = self._conn.execute(q)
         if rows:
-            return [row["id"] for row in rows]
-
-        return []
+            return map(operator.itemgetter(0), rows)
 
     def get_tokens(self, token_ids):
         q = "SELECT id FROM tokens WHERE id IN %s" % \
@@ -591,9 +574,7 @@ class Graph:
 
         rows = self._conn.execute(q)
         if rows:
-            return [row["id"] for row in rows]
-
-        return []
+            return map(operator.itemgetter(0), rows)
 
     def get_node_by_tokens(self, tokens):
         c = self.cursor()
@@ -610,17 +591,12 @@ class Graph:
         c.execute(q, tokens)
         return c.lastrowid
 
-    def get_node_tokens(self, node_id):
-        q = "SELECT %s FROM nodes WHERE id = ?" % self._all_tokens
+    def get_text_by_edge(self, edge_id):
+        q = "SELECT tokens.text, edges.has_space FROM nodes, edges, tokens " \
+            "WHERE edges.id = ? AND edges.prev_node = nodes.id " \
+            "AND nodes.%s = tokens.id" % self._last_token
 
-        row = self._conn.execute(q, (node_id,)).fetchone()
-        assert row is not None
-
-        return tuple(row)
-
-    def get_node_text(self, node_id):
-        tokens = self.get_node_tokens(node_id)
-        return [self.get_token_by_id(token_id) for token_id in tokens]
+        return self._conn.execute(q, (edge_id,)).fetchone()
 
     def get_random_token(self):
         # token 1 is the end_token_id, so we want to generate a random token
@@ -641,6 +617,27 @@ class Graph:
         if row:
             return int(row[0])
 
+    def get_edge_logprob(self, edge_id):
+        # Each edge goes from an n-gram node (word1, word2, word3) to
+        # another (word2, word3, word4). Calculate the probability:
+        # P(word4|word1,word2,word3) = count(edge_id) / count(prev_node_id)
+
+        c = self.cursor()
+        q = "SELECT edges.count, nodes.count FROM edges, nodes " \
+            "WHERE edges.id = ? AND edges.prev_node = nodes.id"
+
+        edge_count, node_count = c.execute(q, (edge_id,)).fetchone()
+        return math.log(edge_count, 2) - math.log(node_count, 2)
+
+    def has_space(self, edge_id):
+        c = self.cursor()
+
+        q = "SELECT has_space FROM edges WHERE id = ?"
+
+        row = c.execute(q, (edge_id,)).fetchone()
+        if row:
+            return bool(row[0])
+
     def add_edge(self, prev_node, next_node, has_space):
         c = self.cursor()
 
@@ -662,43 +659,58 @@ class Graph:
         # incremented here, to register that the node has been seen an
         # additional time. This is now handled by database triggers.
 
-    def get_node_count(self, node_id):
-        q = "SELECT count FROM nodes WHERE nodes.id = ?"
-
-        row = self._conn.execute(q, (node_id,)).fetchone()
-        assert row
-
-        return row[0]
-
-    def get_node_counts(self, node_ids):
-        q = "SELECT id, count FROM nodes WHERE id IN %s" % \
-            self.get_seq_expr(node_ids)
-
-        return self._conn.execute(q)
-
-    def walk(self, node, end_id, direction, append):
-        """Perform a random walk on the graph starting at node"""
+    def search_bfs(self, start_id, end_id, direction):
         if direction:
-            q = "SELECT id, next_node, prev_node, has_space, count " \
+            q = "SELECT id, next_node FROM edges WHERE prev_node = ?"
+        else:
+            q = "SELECT id, prev_node FROM edges WHERE next_node = ?"
+
+        c = self.cursor()
+
+        left = collections.deque([(start_id, tuple())])
+        while left:
+            cur, path = left.popleft()
+            rows = c.execute(q, (cur,))
+
+            for rowid, next in rows:
+                newpath = path + (rowid,)
+
+                if next == end_id:
+                    yield newpath
+                else:
+                    left.append((next, newpath))
+
+    def search_random_walk(self, start_id, end_id, direction):
+        """Walk once randomly from start_id to end_id."""
+        if direction:
+            q = "SELECT id, next_node " \
                 "FROM edges WHERE prev_node = :last " \
                 "LIMIT 1 OFFSET abs(random())%(SELECT count(*) from edges " \
                 "                              WHERE prev_node = :last)"
         else:
-            q = "SELECT id, prev_node, next_node, has_space, count " \
+            q = "SELECT id, prev_node " \
                 "FROM edges WHERE next_node = :last " \
                 "LIMIT 1 OFFSET abs(random())%(SELECT count(*) from edges " \
                 "                              WHERE next_node = :last)"
 
         c = self.cursor()
-        last_node = node
 
-        while last_node != end_id:
-            row = c.execute(q, dict(last=last_node)).fetchone()
+        left = collections.deque([(start_id, tuple())])
+        while left:
+            cur, path = left.popleft()
+            rows = c.execute(q, dict(last=cur))
 
-            append(Edge(self, row["id"], row["prev_node"], row["next_node"],
-                        row["has_space"], row["count"]))
+            # Note: the LIMIT 1 above means this list only contains
+            # one row. Using a list here so this matches the bfs()
+            # code, so the two functions can be more easily combined
+            # later.
+            for rowid, next in rows:
+                newpath = path + (rowid,)
 
-            last_node = row[1]
+                if next == end_id:
+                    yield newpath
+                else:
+                    left.append((next, newpath))
 
     def init(self, order, tokenizer, run_migrations=True):
         c = self.cursor()
@@ -808,10 +820,12 @@ CREATE UNIQUE INDEX IF NOT EXISTS edges_all_prev ON edges
             insert_q = "INSERT INTO token_stems (token_id, stem) VALUES (?, ?)"
 
             q = c.execute("""
-SELECT id, text FROM tokens WHERE is_word = 1""")
+SELECT id, text FROM tokens""")
 
             for row in q:
-                insert_c.execute(insert_q, (row[0], stemmer.stem(row[1])))
+                stem = stemmer.stem(row[1])
+                if stem is not None:
+                    insert_c.execute(insert_q, (row[0], stem))
 
             self.commit()
 
